<?xml version="1.0" encoding="UTF-8" ?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
    <xsl:output method="text" encoding="utf-8"/>
    <!--
      XSL to transform T-Plan Robot workitems or schedules to the Amazon State Language-like 
      JSON format understood by RAP 3.0
      
      Version 5.0 from 17 Nov 2020
        - Fixed "mandatory":"true" to "mandatory":true
      Author: Robert Pes, robert.pes@t-plan.com
    -->
    <xsl:template match="/tprschedule">
        <!-- Create the initial work item header -->
        <xsl:variable name="firstTask" select="testsuite/task[not(./dependency)]/@uid"/>
        <xsl:text>{</xsl:text>
        <xsl:text>"process":"</xsl:text>
        <xsl:call-template name="escape-string">
            <xsl:with-param name="s" select="testsuite/@name"/>
        </xsl:call-template>
        <xsl:text>",</xsl:text>
        <xsl:text>"type":"workitem",</xsl:text>      
        <xsl:if test="testsuite/@subtype">
            <xsl:text>"subtype":"</xsl:text>
            <xsl:call-template name="escape-string">
                <xsl:with-param name="s" select="testsuite/@subtype"/>
            </xsl:call-template>
            <xsl:text>",</xsl:text>
        </xsl:if>
        <xsl:text>"id":"</xsl:text>
        <xsl:call-template name="escape-string">
            <xsl:with-param name="s" select="@uid"/>
        </xsl:call-template>
        <xsl:text>",</xsl:text>
        <xsl:text>"initialStep":"</xsl:text>
        <xsl:call-template name="escape-string">
            <xsl:with-param name="s" select="$firstTask"/>
        </xsl:call-template>
        <xsl:text>",</xsl:text>
        <xsl:text>"comments":"",</xsl:text>
        <xsl:text>"param":{</xsl:text>
        <xsl:apply-templates select="testsuite"/>
        <xsl:text>}</xsl:text>
    </xsl:template>

    <xsl:template match="testsuite">
        <!-- Code the input parameters -->
        <xsl:text>"input":[</xsl:text>
        <xsl:apply-templates select="input"/>
        <xsl:text>],</xsl:text>

        <!-- Code the output parameters -->
        <xsl:text>"output":[</xsl:text>
        <!-- RAP3-87: Improve handling of misconfigured angents -->
        <xsl:text>{"name":"_RAP_EXITCODE","type":"int","label":"Workitem exit code"},{"name":"_RAP_EXITREASON","type":"string","label":"Workitem exit reason"},{"name":"_RAP_EXITSTATUS","type":"string","label":"Workitem exit status"}</xsl:text>
        <xsl:if test="output/param">
            <xsl:text>,</xsl:text>
        </xsl:if>
        <xsl:apply-templates select="output"/>
        <xsl:text>]</xsl:text>
        <xsl:text>},</xsl:text>
  
        <!-- Code the steps, i.e. the scripts and the choices -->
        <xsl:text>"steps":{</xsl:text>
        <xsl:for-each select="task">
            <xsl:text>"</xsl:text>
            <xsl:value-of select="@uid"/>
            <xsl:text>":{</xsl:text>
            <xsl:text>"type":"Script",</xsl:text>
            <xsl:text>"comment":"</xsl:text>
            <xsl:call-template name="escape-string">
                <xsl:with-param name="s" select="script"/>
            </xsl:call-template>
            <xsl:text>",</xsl:text>
            <!-- This handles script results injected into the ML -->
            <xsl:if test="@result">
                <xsl:text>"results":{"status":"</xsl:text>
                <xsl:value-of select="@result"/>
                <xsl:text>","execTime":"</xsl:text>
                <xsl:value-of select="@execTime"/>
                <xsl:text>"</xsl:text>
                <xsl:if test="@info">
                    <xsl:text>","info":"</xsl:text>
                    <xsl:value-of select="@info"/>
                    <xsl:text>"</xsl:text>
                </xsl:if>
                <xsl:text>},</xsl:text>
            </xsl:if>
            <xsl:text>"label":"</xsl:text>
            <!-- Populate "comments" with either the name or the task file name -->
            <xsl:choose>
                <xsl:when test="string-length(@name) = 0"><!-- The task name is empty -> use file name without extension -->                 
                    <xsl:choose>
                        <xsl:when test="contains(script, '\')"><!-- The path contains '\' -> use Windows path extraction -->                        
                            <xsl:variable name="lw">
                                <xsl:call-template name="extract-filename-windows">
                                    <xsl:with-param name="path" select="script"/>
                                </xsl:call-template>
                            </xsl:variable>
                            <xsl:choose>
                                <xsl:when test="contains($lw, '.')">                           
                                    <xsl:value-of select="substring-before($lw, '.')"/>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:value-of select="$lw"/>
                                </xsl:otherwise>
                            </xsl:choose>
                        </xsl:when>
                        <xsl:otherwise><!-- Fall back to Linux/Unix path extraction -->   
                            <xsl:variable name="lu">
                                <xsl:call-template name="extract-filename-linux">
                                    <xsl:with-param name="path" select="script"/>
                                </xsl:call-template>
                            </xsl:variable>
                            <xsl:choose>
                                <xsl:when test="contains($lu, '.')">                           
                                    <xsl:value-of select="substring-before($lu, '.')"/>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:value-of select="$lu"/>
                                </xsl:otherwise>
                            </xsl:choose>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:when>
                <xsl:otherwise>
                    <!-- <xsl:value-of select="@name"/> -->
                    <xsl:call-template name="escape-string">
                        <xsl:with-param name="s" select="@name"/>
                    </xsl:call-template>
                </xsl:otherwise>
            </xsl:choose>
            <xsl:text>",</xsl:text>
      
            <!-- Identify if there are any next scripts depending on this one -->
            <xsl:variable name="tuid" select="@uid"/>
            <xsl:variable name="deps" select="parent::testsuite/task[@type='internal']/dependencies/dependency[@uid=$tuid]"/>
            <xsl:variable name="depsfinish" select="parent::testsuite/task[@type='internal']/dependencies/dependency[(@uid=$tuid) and (@type='finish')]"/>

            <xsl:choose>
                <xsl:when test="$deps">
                    <xsl:choose>
                        <xsl:when test="count($deps) = 1 and count($depsfinish) = 1">
                            <xsl:text>"next":"</xsl:text>
                            <xsl:value-of select="$depsfinish/../../@uid"/>
                            <xsl:text>",</xsl:text>
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:text>"next":"choice</xsl:text>
                            <xsl:value-of select="$tuid"/>
                            <xsl:text>",</xsl:text>
                            <xsl:text>"defaultStep":"EndChoice",</xsl:text>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:text>"next":"EndChoice",</xsl:text>
                </xsl:otherwise>
            </xsl:choose>
      
            <xsl:text>"id":"</xsl:text>
            <xsl:value-of select="@uid"/>
            <xsl:text>"</xsl:text>
            <xsl:text>}</xsl:text>
      
            <xsl:choose>
                <xsl:when test="count($deps) = 1 and count($depsfinish) = 1">
                    <!-- There's just one dependency and it's 'Finish' - no choice is needed -->
                </xsl:when>
                <xsl:when test="$deps">
                    <!-- There are next scripts -> create the Choice object -->
                    <xsl:text>,"choice</xsl:text>
                    <xsl:value-of select="$tuid"/>
                    <xsl:text>":{</xsl:text>
                    <xsl:text>"comment":"Pass or Fail?","type":"Choice",</xsl:text>
                    <xsl:text>"id":"choice</xsl:text>
                    <xsl:value-of select="$tuid"/>
                    <xsl:text>",</xsl:text>
                    <xsl:text>"choices":[</xsl:text>
        
                    <xsl:variable name="pass" select="parent::testsuite/task[@type='internal']/dependencies/dependency[(@uid=$tuid) and (@type='pass')]"/>
                    <xsl:variable name="fail" select="parent::testsuite/task[@type='internal']/dependencies/dependency[(@uid=$tuid) and (@type='fail')]"/>
                    <xsl:variable name="finish" select="parent::testsuite/task[@type='internal']/dependencies/dependency[(@uid=$tuid) and (@type='finish')]"/>

                    <!-- Code the Pass dependency type -->
                    <xsl:if test="$pass">
                        <xsl:text>{"variable":"EXITSTATUS","operation":"StringEquals","operand":"Pass",</xsl:text>
                        <xsl:text>"next":"</xsl:text>
                        <xsl:value-of select="$pass[1]/ancestor::task/@uid"/>
                        <xsl:text>"</xsl:text>
                        <xsl:text>}</xsl:text> <!-- End of "Pass" -->
                    </xsl:if>
        
                    <!-- Code the Fail dependency type -->
                    <xsl:if test="$fail">
                        <xsl:if test="$pass">
                            <xsl:text>,</xsl:text>
                        </xsl:if>
                        <xsl:text>{"variable":"EXITSTATUS","operation":"StringEquals","operand":"Fail",</xsl:text>
                        <xsl:text>"next":"</xsl:text>
                        <xsl:value-of select="$fail[1]/ancestor::task/@uid"/>
                        <xsl:text>"</xsl:text>
                        <xsl:text>}</xsl:text> <!-- End of "Fail" -->
                    </xsl:if>
        
                    <!-- Code the Finish dependency type -->
                    <!-- NOT SUPPORTED ATM BECAUSE IN A COMBINATION WITH A PASS+FAIL DEPENDENCY IT CREATES A PARALLEL BRANCH -->
                    <!--
                    <xsl:if test="$finish">
                      <xsl:if test="$pass or $fail">
                        <xsl:text>,</xsl:text>
                      </xsl:if>
                      <xsl:text>{"variable":"0","operation":"StringEquals","operand":"Any",</xsl:text>
                      <xsl:text>"next":"</xsl:text><xsl:value-of select="$finish[1]/ancestor::task/@uid"/><xsl:text>"</xsl:text>
                      <xsl:text>}</xsl:text>
                    </xsl:if>
                    -->
        
                    <!-- Close the Choice object -->
                    <xsl:text>]</xsl:text> <!-- end of 'choices' -->
                    <xsl:text>}</xsl:text>
                </xsl:when>
            </xsl:choose>

            <xsl:if test="position() != last()">
                <xsl:text>,</xsl:text>
            </xsl:if>
        </xsl:for-each>
        <!-- This is to prevent adding a comma when there are no tasks -->
        <xsl:if test="task">
            <xsl:text>,</xsl:text>
        </xsl:if>
        <xsl:text>"EndChoice":{"comment":"Pass or Fail?","type":"Choice","id":"EndChoice","defaultStep":"DefaultExit","choices":[{"variable":"EXITSTATUS","operation":"StringEquals","operand":"Fail","next":"FailExit"}]}</xsl:text>    
        <xsl:text>,"DefaultExit":{"comment":"End","id":"DefaultExit","type":"Success"},"FailExit":{"comment":"Terminated","id":"FailExit","type":"Failure"}</xsl:text>
        <xsl:text>}</xsl:text>
    </xsl:template>
    
    <xsl:template name="extract-filename-linux">
        <xsl:param name="path" select="." />
        <xsl:choose>
            <xsl:when test="not(contains($path, '\'))">
                <xsl:value-of select="$path" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:call-template name="extract-filename-linux">
                    <xsl:with-param name="path" select="substring-after($path, '\')" />
                </xsl:call-template>
            </xsl:otherwise>
        </xsl:choose>    
    </xsl:template>
    
    <xsl:template name="extract-filename-windows">
        <xsl:param name="path" select="." />
        <xsl:choose>
            <xsl:when test="not(contains($path, '\'))">
                <xsl:value-of select="$path" />
            </xsl:when>
            <xsl:otherwise>
                <xsl:call-template name="extract-filename-windows">
                    <xsl:with-param name="path" select="substring-after($path, '\')" />
                </xsl:call-template>
            </xsl:otherwise>
        </xsl:choose>    
    </xsl:template>

    <xsl:template match="input">
        <xsl:apply-templates select="param"/>
    </xsl:template>

    <xsl:template match="output">
        <xsl:apply-templates select="param"/>
    </xsl:template>

    <!-- Template to create a single input or output parameter JSON -->
    <xsl:template match="param">
        <xsl:text>{</xsl:text>
        <xsl:text>"name":"</xsl:text>
        <xsl:call-template name="escape-string">
            <xsl:with-param name="s" select="name"/>
        </xsl:call-template>
        <xsl:text>",</xsl:text>
        <xsl:if test="./desc">
            <xsl:text>"desc":"</xsl:text>
            <xsl:call-template name="escape-string">
                <xsl:with-param name="s" select="desc"/>
            </xsl:call-template>
            <xsl:text>",</xsl:text>
        </xsl:if>
        <xsl:if test="./mandatory">
            <xsl:text>"mandatory":</xsl:text>
            <xsl:call-template name="escape-string">
                <xsl:with-param name="s" select="mandatory"/>
            </xsl:call-template>
            <xsl:text>,</xsl:text>
        </xsl:if>
        <xsl:if test="./type">
            <xsl:text>"type":"</xsl:text>
            <xsl:value-of select="type"/>
            <xsl:text>",</xsl:text>
            <xsl:choose>
                <!-- For 'select' we support the 'options' -->
                <xsl:when test="type = 'select'">
                    <xsl:if test="options">
                        <xsl:text>"options":[</xsl:text>
                        <xsl:for-each select="options/option">
                            <xsl:text>{"option":"</xsl:text>
                            <xsl:call-template name="escape-string">
                                <xsl:with-param name="s" select="."/>
                            </xsl:call-template>
                            <xsl:text>"}</xsl:text>
                            <xsl:if test="position() != last()">
                                <xsl:text>,</xsl:text>
                            </xsl:if>
                        </xsl:for-each>
                        <xsl:text>],</xsl:text>
                    </xsl:if>                                   
                    <xsl:if test="./value">
                        <xsl:text>"value":"</xsl:text>
                        <xsl:call-template name="escape-string">
                            <xsl:with-param name="s" select="value"/>
                        </xsl:call-template>
                        <xsl:text>",</xsl:text>
                    </xsl:if>
                </xsl:when>
                <!-- For 'int' and 'float' we support the optional 'min' and 'max' attributes -->
                <xsl:when test="(type='int') or (type='float')">
                    <xsl:if test="min">
                        <xsl:text>"min":</xsl:text>
                        <xsl:value-of select="min"/>
                        <xsl:text>,</xsl:text>
                    </xsl:if>                                   
                    <xsl:if test="max">
                        <xsl:text>"max":</xsl:text>
                        <xsl:value-of select="max"/>
                        <xsl:text>,</xsl:text>
                    </xsl:if>                                   
                    <xsl:if test="./value">
                        <xsl:text>"value":</xsl:text>
                        <xsl:value-of select="value"/>
                        <xsl:text>,</xsl:text>
                    </xsl:if>
                </xsl:when>
                <!-- For other nodes output the value as string -->
                <xsl:otherwise>
                    <xsl:if test="./value">
                        <xsl:text>"value":"</xsl:text>
                        <xsl:call-template name="escape-string">
                            <xsl:with-param name="s" select="value"/>
                        </xsl:call-template>
                        <xsl:text>",</xsl:text>
                    </xsl:if>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:if>
        <xsl:text>"label":"</xsl:text>
        <xsl:call-template name="escape-string">
            <xsl:with-param name="s" select="label"/>
        </xsl:call-template>
        <xsl:text>"</xsl:text>
        <xsl:text>}</xsl:text>
         <xsl:if test="following-sibling::*"><xsl:text>,</xsl:text></xsl:if> 
    </xsl:template>

    <xsl:template name="escape-string">
        <xsl:param name="s"/>
        <!-- <xsl:text>"</xsl:text> -->
        <xsl:call-template name="escape-bs-string">
            <xsl:with-param name="s" select="$s"/>
        </xsl:call-template>
        <!-- <xsl:text>"</xsl:text> -->
    </xsl:template>


    <!-- Escape the backslash (\) before everything else. -->
    <xsl:template name="escape-bs-string">
        <xsl:param name="s"/>
        <xsl:choose>
            <xsl:when test="contains($s,'\')">
                <xsl:call-template name="escape-quot-string">
                    <xsl:with-param name="s" select="concat(substring-before($s,'\'),'\\')"/>
                </xsl:call-template>
                <xsl:call-template name="escape-bs-string">
                    <xsl:with-param name="s" select="substring-after($s,'\')"/>
                </xsl:call-template>
            </xsl:when>
            <xsl:otherwise>
                <xsl:call-template name="escape-quot-string">
                    <xsl:with-param name="s" select="$s"/>
                </xsl:call-template>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>


    <!-- Escape the double quote ("). -->
    <xsl:template name="escape-quot-string">
        <xsl:param name="s"/>
        <xsl:choose>
            <xsl:when test="contains($s,'&quot;')">
                <xsl:call-template name="encode-string">
                    <xsl:with-param name="s" select="concat(substring-before($s,'&quot;'),'\&quot;')"/>
                </xsl:call-template>
                <xsl:call-template name="escape-quot-string">
                    <xsl:with-param name="s" select="substring-after($s,'&quot;')"/>
                </xsl:call-template>
            </xsl:when>
            <xsl:otherwise>
                <xsl:call-template name="encode-string">
                    <xsl:with-param name="s" select="$s"/>
                </xsl:call-template>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>


    <!-- Replace tab, line feed and/or carriage return by its matching escape code. Can't escape backslash
    or double quote here, because they don't replace characters (&#x0; becomes \t), but they prefix
    characters (\ becomes \\). Besides, backslash should be seperate anyway, because it should be
    processed first. This function can't do that. -->
    <xsl:template name="encode-string">
        <xsl:param name="s"/>
        <xsl:choose>
            <!-- tab -->
            <xsl:when test="contains($s,'&#x9;')">
                <xsl:call-template name="encode-string">
                    <xsl:with-param name="s" select="concat(substring-before($s,'&#x9;'),'\t',substring-after($s,'&#x9;'))"/>
                </xsl:call-template>
            </xsl:when>
            <!-- line feed -->
            <xsl:when test="contains($s,'&#xA;')">
                <xsl:call-template name="encode-string">
                    <xsl:with-param name="s" select="concat(substring-before($s,'&#xA;'),'\n',substring-after($s,'&#xA;'))"/>
                </xsl:call-template>
            </xsl:when>
            <!-- carriage return -->
            <xsl:when test="contains($s,'&#xD;')">
                <xsl:call-template name="encode-string">
                    <xsl:with-param name="s" select="concat(substring-before($s,'&#xD;'),'\r',substring-after($s,'&#xD;'))"/>
                </xsl:call-template>
            </xsl:when>
            <xsl:otherwise>
                <xsl:value-of select="$s"/>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>


    <!-- number (no support for javascript mantise) -->
    <xsl:template match="text()[not(string(number())='NaN')]" mode="xml-to-json" >
        <xsl:value-of select="."/>
    </xsl:template>


    <!-- boolean, case-insensitive -->
    <xsl:template match="text()[translate(.,'TRUE','true')='true']" mode="xml-to-json" >true</xsl:template>
    <xsl:template match="text()[translate(.,'FALSE','false')='false']" mode="xml-to-json" >false</xsl:template>


</xsl:stylesheet>